package no.vestein.rtree.lsm;

import org.davidmoten.hilbert.HilbertCurve;
import org.davidmoten.hilbert.SmallHilbertCurve;

import java.util.concurrent.atomic.AtomicInteger;

public class BitMagic {
  private static long x = System.currentTimeMillis();
  public static AtomicInteger COUNTER = new AtomicInteger();

  public static long randomLong() {
    x ^= (x << 21);
    x ^= (x >>> 35);
    x ^= (x << 4);
    return x;
  }

  public static interface SpaceCurve {
    long encode(float x, float y);
    long encode(int x, int y);
    String shortName();
  }

  /**
   * http://graphics.stanford.edu/~seander/bithacks.html#InterleaveBMN
   */
  public static class ZOrder implements SpaceCurve {

    private static final long[] B = {
        0x5555555555555555L,
        0x3333333333333333L,
        0x0f0f0f0f0f0f0f0fL,
        0x00ff00ff00ff00ffL,
        0x0000ffff0000ffffL
    };

    private static final int[] S = {1, 2, 4, 8, 16};

    private long encode1(long x, long y) {
      x = (x | (x << S[4])) & B[4];
      x = (x | (x << S[3])) & B[3];
      x = (x | (x << S[2])) & B[2];
      x = (x | (x << S[1])) & B[1];
      x = (x | (x << S[0])) & B[0];

      y = (y | (y << S[4])) & B[4];
      y = (y | (y << S[3])) & B[3];
      y = (y | (y << S[2])) & B[2];
      y = (y | (y << S[1])) & B[1];
      y = (y | (y << S[0])) & B[0];

      return x | (y << 1);
    }

    /**
     * Assumes X and Y is below 2^9
     */
    @Override
    public long encode(float x, float y) {
      int xx = (int) ((x + 90.0) * 0x0080000);
      int yy = (int) ((y + 180) * 0x0080000);
      return encode(xx, yy);
    }

    @Override
    public long encode(int x, int y) {
      return encode1(x, y);
    }

    @Override
    public String shortName() {
      return "Z";
    }

  }

  /**
   * https://github.com/davidmoten/hilbert-curve
   */
  public static class Hilbert implements SpaceCurve {
    SmallHilbertCurve curve;

    public Hilbert() {
      this(31, 2);
    }

    public Hilbert(int bits, int dimensions) {
      curve = HilbertCurve.small().bits(bits).dimensions(dimensions);
    }

    @Override
    public long encode(float x, float y) {
      int xx = (int) ((x + 90.0) * 0x0080000);
      int yy = (int) ((y + 180) * 0x0080000);
      return encode(xx, yy);
    }

    @Override
    public long encode(int x, int y) {
      return curve.index(x, y);
//      return encode1(x, y, this.r);
    }

    @Override
    public String shortName() {
      return "H";
    }

  }

}
