package no.vestein.rtree.lsm;

import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

public class ComponentManager {
  public final int threshold;
  public final int fanout;
  public final int level;
  public final AtomicBoolean merging;
  public PersistentTree[] trees;
  private final String dataDir;
  private final ExecutorService pool;
  private AtomicInteger generation;
  public AtomicInteger runCounter;

  public ComponentManager(int fanout, int level, int threshold, ExecutorService pool, String dataDir, LSM.MergeStrat mergeStrat) {
    this.threshold = threshold;
    this.fanout = fanout;
    this.level = level;
    this.pool = pool;
    this.dataDir = dataDir;

    runCounter = new AtomicInteger();
    merging = new AtomicBoolean(false);
    trees = new PersistentTree[threshold];
    generation = new AtomicInteger();

    IOLSM.makeGenerationFolder(dataDir, level, generation.get());

    for (int i = 0; i < threshold; i++) {
      trees[i] = mergeStrat.createInstance(fanout, generation.get(), level, i, dataDir);
    }
  }

  public void updateGeneration() {
    int g = generation.incrementAndGet();
    PersistentTree tree;
    for (int i = 0; i < threshold; i++) {
      tree = trees[i];
      tree.generation = g;
    }
    IOLSM.makeGenerationFolder(dataDir, level, g);
  }

  public boolean addRun(int id, List<Entry> entries, LSM.MergeTaskTree task) {
    int currentRuns = runCounter.getAndUpdate(c -> {
      if (c == threshold) return c;
      return ++c;
    });
    if (currentRuns >= threshold) {
      if (merging.compareAndSet(false, true)) {
        pool.submit(task);
      } else {
        runCounter.getAndDecrement();
      }
      return false;
    }

    trees[currentRuns].createTree(entries);
    return true;
  }

  public PersistentTree getFreeRun(LSM.MergeTaskTree task) {
    if (!merging.get()) {
//      int currentRuns = runCounter.getAndAdd(1);
      int currentRuns = runCounter.getAndUpdate(c -> {
        if (c == threshold) return c;
        return ++c;
      });
      if (currentRuns == threshold) {
        // TODO: do merge
        if (merging.compareAndSet(false, true)) {
          pool.submit(task);
        }

        return null;
      }

      return trees[currentRuns];
    }
    return null;
  }

}
