package no.vestein.rtree.lsm;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.util.LinkedList;

public class IOLSM {
  public static int ENTRY_SIZE = 32;
  public static final int NODE_SIZE = 36;
  public static final int HEADER_SIZE = 17;
  public static int READ_BUFFER_SIZE = 0;
  public static int WRITE_BUFFER_SIZE = 0;


  public static int fillBufferMapped(LinkedList<Entry> entryBuffer, ByteBuffer buffer, byte[] b) throws IOException {
    while (buffer.remaining() > 0 && entryBuffer.size() < 10000) {
      buffer.get(b);
      entryBuffer.add(IOLSM.bytesToEntry(b));
    }

    int st = 0;
    if (buffer.remaining() > 0) {
      st = 1;
    }

    return st;
  }

  public static int fillBufferNodes1(LinkedList<Node> nodeBuffer, FileInputStream stream, ByteBuffer buffer) throws IOException {
    byte[] b = new byte[IOLSM.NODE_SIZE];
    FileChannel fc = stream.getChannel();

    if (buffer.remaining() > fc.size() - fc.position()) {
      buffer.limit((int) (fc.size() - fc.position()));
    }

    int st = fc.read(buffer);

    buffer.flip();
    while(buffer.remaining() >= IOLSM.NODE_SIZE) {
      buffer.get(b);
      nodeBuffer.add(IOLSM.bytesToNode(b));
    }
    buffer.compact();
    return st;
  }

  public static int fillBuffer1(LinkedList<Entry> entryBuffer, FileInputStream stream, ByteBuffer buffer) throws IOException {
    byte[] b = new byte[IOLSM.ENTRY_SIZE];
    FileChannel fc = stream.getChannel();

    if (buffer.remaining() > fc.size() - fc.position()) {
      buffer.limit((int) (fc.size() - fc.position()));
    }

    int st = fc.read(buffer);

    buffer.flip();
    while(buffer.remaining() >= IOLSM.ENTRY_SIZE) {
      buffer.get(b);
      entryBuffer.add(IOLSM.bytesToEntry(b));
    }

    buffer.compact();
    return st;
  }

  public static void makeGenerationFolder(String dataDir, int level, int generation) {
    File dir = new File(String.format("%sL%d/%d", dataDir, level, generation));
    dir.mkdirs();
  }

  public static byte[] entryToBytes1(Entry entry, ByteBuffer buffer) {
    buffer.position(0);
    buffer.putFloat(entry.xl);
    buffer.putFloat(entry.xh);
    buffer.putFloat(entry.yl);
    buffer.putFloat(entry.yh);
    buffer.putLong(entry.lhv);
    buffer.putLong(entry.uuid);

    return buffer.array();
  }

  public static byte[] entryToBytes(Entry entry) {
    ByteBuffer buffer = ByteBuffer.allocate(ENTRY_SIZE);

    buffer.putFloat(entry.xl);
    buffer.putFloat(entry.xh);
    buffer.putFloat(entry.yl);
    buffer.putFloat(entry.yh);
    buffer.putLong(entry.lhv);
    buffer.putLong(entry.uuid);

    return buffer.array();
  }

  public static Entry bytesToEntry(byte[] bytes) {
    ByteBuffer buffer = ByteBuffer.wrap(bytes);

    return new Entry(
        buffer.getFloat(),
        buffer.getFloat(),
        buffer.getFloat(),
        buffer.getFloat(),
        buffer.getLong(),
        buffer.getLong()
    );

  }

  public static byte[] nodeToBytes(Node node) {
    ByteBuffer buffer = ByteBuffer.allocate(NODE_SIZE);

    buffer.putFloat(node.xl);
    buffer.putFloat(node.xh);
    buffer.putFloat(node.yl);
    buffer.putFloat(node.yh);
    buffer.putLong(node.lhv);
    buffer.putInt(node.start);
    buffer.putInt(node.length);
    buffer.putInt(node.level);

    return buffer.array();
  }

  public static Node bytesToNode(byte[] bytes) {
    ByteBuffer buffer = ByteBuffer.wrap(bytes);

    return new Node(
        buffer.getFloat(),
        buffer.getFloat(),
        buffer.getFloat(),
        buffer.getFloat(),
        buffer.getLong(),
        buffer.getInt(),
        buffer.getInt(),
        buffer.getInt()
    );
  }

  public static byte[] nodeFileHeaderToBytes(NodeFileHeader header) {
    ByteBuffer buffer = ByteBuffer.allocate(HEADER_SIZE);

    buffer.putLong(header.size);
    buffer.putInt(header.elementLength);
    buffer.putInt(header.level);
    buffer.put(header.type);

    return buffer.array();
  }

  public static NodeFileHeader bytesToNodeFileHeader(byte[] bytes) {
    ByteBuffer buffer = ByteBuffer.wrap(bytes);

    return new NodeFileHeader(
        buffer.getLong(),
        buffer.getInt(),
        buffer.getInt(),
        buffer.get()
    );
  }

  public static class NodeFileHeader {
    public long size;
    public int elementLength;
    public int level;
    public byte type;

    public static final byte ENTRY = 0;
    public static final byte NODE = 1;
    public static final byte ROOT = 2;

    public NodeFileHeader(long size, int elementLength, int level, byte type) {
      this.size = size;
      this.elementLength = elementLength;
      this.level = level;
      this.type = type;
    }

  }

}
