package no.vestein.rtree.lsm;

import no.vestein.rtree.pre.ZRTree;

import javax.imageio.ImageIO;
import java.awt.*;
import java.awt.geom.AffineTransform;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;

/**
 * Created by veste
 * Date: 09.04.2019
 * Time: 00:38
 */
public class Visuals {
  private final static float dash1[] = {10.0f};
  private final static BasicStroke dashed = new BasicStroke(1.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash1, 0.0f);
  private final static BasicStroke dashed2 = new BasicStroke(2.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash1, 0.0f);
  private final static BasicStroke dashed3 = new BasicStroke(4.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash1, 0.0f);
  private final static BasicStroke dashed4 = new BasicStroke(8.0f, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, dash1, 0.0f);
  private final static BasicStroke normal = new BasicStroke(1.0f);
  private final static BasicStroke normal2 = new BasicStroke(2.0f);
  private final static BasicStroke normal3 = new BasicStroke(4.0f);
  private final static BasicStroke normal4 = new BasicStroke(8.0f);

  private static void drawLeafLevel(Graphics2D g2d, AffineTransform at, PersistentTree tree, int level, int scale, float modifier) {
    try {

      LinkedList<Entry> entryBuffer = new LinkedList<>();
      FileInputStream stream = new FileInputStream(tree.getLeafFile());
      ByteBuffer buffer = ByteBuffer.allocate((int) Math.pow(2, 16));

      IOLSM.NodeFileHeader header = IOLSM.bytesToNodeFileHeader(stream.readNBytes(IOLSM.HEADER_SIZE));

      IOLSM.fillBuffer1(entryBuffer, stream, buffer);
      while (!entryBuffer.isEmpty()) {

        Entry e = entryBuffer.pollFirst();

        float x = (e.yl + 180.0f + 5.0f) * scale;
        float y = (e.xl + 90.0f + 5.0f) * scale;

        g2d.draw(new Rectangle2D.Float(x - 1, y - 1, 2, 2));
//        g2d.draw(at.createTransformedShape(new Rectangle2D.Float(x, y, 1, 1)));

        if (entryBuffer.isEmpty()) {
          IOLSM.fillBuffer1(entryBuffer, stream, buffer);
        }
      }

      stream.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private static void drawNodeLevel(Graphics2D g2d, AffineTransform at, PersistentTree tree, int level, int scale, float modifier) {
    try {
      g2d.setComposite(AlphaComposite.getInstance(3, 0.10f));
      g2d.setColor(Color.BLUE);

      LinkedList<Node> entryBuffer = new LinkedList<>();
      FileInputStream stream = new FileInputStream(tree.getLevelFileName(level));
      ByteBuffer buffer = ByteBuffer.allocate((int) Math.pow(2, 16));

      IOLSM.NodeFileHeader header = IOLSM.bytesToNodeFileHeader(stream.readNBytes(IOLSM.HEADER_SIZE));
      int c = 0;

      IOLSM.fillBufferNodes1(entryBuffer, stream, buffer);
      while (!entryBuffer.isEmpty()) {

        Node e = entryBuffer.pollFirst();
        c++;

        float x = (e.yl + 180.0f + 5.0f) * scale;
        float y = (e.xl + 90.0f + 5.0f) * scale;
        float width = (e.yh - e.yl) * scale;
        float height = (e.xh - e.xl) * scale;

        g2d.fill(new Rectangle2D.Float(x, y, width, height));
//        g2d.draw(at.createTransformedShape(new Rectangle2D.Float(x, y, width, height)));

        if (entryBuffer.isEmpty()) {
//          IOLSM.fillNodeBuffer(entryBuffer, stream, buffer, remaining);
          IOLSM.fillBufferNodes1(entryBuffer, stream, buffer);
        }
      }

      stream.close();
    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  public static void drawPersisTree(PersistentTree tree, String outputFile) {
    System.out.println("Drawing tree...");
    int scale = 10;
    float modifier = 5.0f;
    int width = 360 + 10;
    int height = 180 + 10;

    BufferedImage image = new BufferedImage(width * scale, height * scale, BufferedImage.TYPE_INT_ARGB);
    Graphics2D g2d = image.createGraphics();

    g2d.setColor(Color.WHITE);
    g2d.fillRect(0, 0, width * scale, height * scale);
    g2d.setStroke(normal);

    AffineTransform at = AffineTransform.getTranslateInstance(width / 2, height / 2);

    g2d.setColor(Color.MAGENTA);
    try {

//      drawLeafLevel(g2d, at, tree, tree.getLeafLevel(), scale, modifier);
      drawNodeLevel(g2d, at, tree, tree.getLeafLevel() - 1, scale, modifier);

      g2d.dispose();

      File file = new File(outputFile);
      ImageIO.write(image, "png", file);

    } catch (IOException e) {
      e.printStackTrace();
    }
    System.out.println("DONE");
  }

  public static void drawZRTree(ZRTree tree, String outputFile, int max, List<Integer> drawlist) {
    System.out.println("Drawing...");

    final int delta = 50;

    BufferedImage image = new BufferedImage(max + (delta * 2), max + (delta * 2), BufferedImage.TYPE_INT_RGB);
    Graphics2D g2d = image.createGraphics();

    g2d.setColor(Color.WHITE);
    g2d.fillRect(0, 0, max + (delta * 2),  max + (delta * 2));
    g2d.setStroke(normal);

    Queue<QueueElement> qu = new LinkedList<>();
    qu.add(new QueueElement(tree.getRoot(), 0));

    LinkedList<QueueElement> list = new LinkedList<>();

    while (!qu.isEmpty()) {
      QueueElement e = qu.remove();
      QueueElement.addChildrenToQueue(qu, e);
      list.add(e);
    }

    while(!list.isEmpty()) {
      QueueElement e = list.removeLast();

      no.vestein.rtree.pre.Node n = e.node;
//      QueueElement.addChildrenToQueue(qu, e);
//      qu.addAll(n.children);
      if (!drawlist.contains(e.level)) continue;

      switch (e.level) {
        case 0:
          g2d.setColor(Color.MAGENTA);
          g2d.setStroke(dashed4);
          break;
        case 1:
          g2d.setColor(Color.GREEN);
          g2d.setStroke(dashed3);
          break;
        case 2:
          g2d.setColor(Color.RED);
          g2d.setStroke(normal);
          break;
        case 3:
          g2d.setColor(Color.CYAN);
          g2d.setStroke(normal);
          break;
        case 4:
          g2d.setColor(Color.MAGENTA);
          g2d.setStroke(normal);
          break;
        default:
          g2d.setColor(Color.BLUE);
          g2d.setStroke(normal);
      }
//      g2d.setColor(Color.MAGENTA);
//      if (n.isLeaf) {
//        g2d.setColor(Color.BLUE);
//      }
      if (n.isLeaf) {
//        g2d.draw(new Rectangle2D.Float(n.xl + 1 + delta, n.yl + 1 + delta, n.xh - n.xl - 1 + delta, n.yh - n.yl - 1 + delta));
        g2d.draw(new Rectangle2D.Float(n.xl + 1 + delta, n.yl + 1 + delta, n.xh - n.xl - 1, n.yh - n.yl - 1));
      } else {
        g2d.draw(new Rectangle2D.Float(n.xl + delta, n.yl + delta, n.xh - n.xl, n.yh - n.yl));
//        g2d.draw(new Rectangle2D.Float(n.xl, n.yl, n.xh - n.xl, n.yh - n.yl));
      }
    }

    g2d.dispose();

    try {
      File file = new File(outputFile);
      ImageIO.write(image, "png", file);
    } catch (IOException e) {

    }

  }

  private static class QueueElement {
    public final no.vestein.rtree.pre.Node node;
    public final int level;
    public QueueElement(no.vestein.rtree.pre.Node node, int level) {
      this.node = node;
      this.level = level;
    }
    public static void addChildrenToQueue(Queue<QueueElement> queue, QueueElement parent) {
      no.vestein.rtree.pre.Node parentNode = parent.node;
      int level = parent.level;
      for (no.vestein.rtree.pre.Node child : parent.node.children) {
        queue.add(new QueueElement(child, level + 1));
      }
    }
  }

}
