package no.vestein.rtree.lsm.cuda;


import jcuda.Pointer;
import jcuda.Sizeof;
import jcuda.jcurand.JCurand;
import jcuda.jcurand.curandGenerator;
import jcuda.runtime.JCuda;

import static jcuda.jcurand.JCurand.*;
import static jcuda.jcurand.curandRngType.CURAND_RNG_PSEUDO_DEFAULT;
import static jcuda.runtime.JCuda.*;
import static jcuda.runtime.cudaMemcpyKind.cudaMemcpyDeviceToHost;

public class CRandom {

  public static float[] generateUniformRandomFloats(int n, long seed) {

    JCuda.setExceptionsEnabled(true);
    JCurand.setExceptionsEnabled(true);

    curandGenerator generator = new curandGenerator();

    float[] hostData = new float[n];

    Pointer deviceData = new Pointer();
    cudaMalloc(deviceData, n * Sizeof.FLOAT);

    curandCreateGenerator(generator, CURAND_RNG_PSEUDO_DEFAULT);

    curandSetPseudoRandomGeneratorSeed(generator, seed);

    curandGenerateUniform(generator, deviceData, n);
//    curandGenerateNormal(generator, deviceData, n, 0, 1.0f);

    cudaMemcpy(Pointer.to(hostData), deviceData, n * Sizeof.FLOAT, cudaMemcpyDeviceToHost);

    curandDestroyGenerator(generator);
    cudaFree(deviceData);

    return hostData;
  }

}
