package no.vestein.rtree.pre;

import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Comparator;
import java.util.TreeSet;

public class IO {
  public static final Comparator<Node> comparator = new Comparator<Node>() {
    @Override
    public int compare(Node o1, Node o2) {
      return Integer.compare(o1.lhv, o2.lhv);
    }
  };

  private static final String DIR = "data/";

  public Node readRoot() {

    byte[] b = readFile( "_root_");
    ByteBuffer buf = ByteBuffer.wrap(b);

    int xl = buf.getInt();
    int xh = buf.getInt();
    int yl = buf.getInt();
    int yh = buf.getInt();
    int lhv = buf.getInt();
    boolean isLeaf = buf.getShort() > 0;

    byte[] sb = new byte[42];
    buf.get(sb, 0, 42);
    String childrenFile = new String(sb, StandardCharsets.UTF_8);
//    String childrenFile = new String(buf.slice().array(), StandardCharsets.UTF_8);

    Node root = new Node(xl, yl, xh, yh, isLeaf);
    root.childrenFilePath = childrenFile.replace("\0", "");
    root.lhv = lhv;

    return root;
  }

  public boolean writeChildren(Node node) {

    byte[] byteChildren = new byte[node.children.size() * 64];

    byte[] buf = new byte[64];

    int i = 0;
    for (Node child : node.children) {
      buf = nodeToByte(child);
      for (int j = 0; j < 64; j++) {
        byteChildren[i + j] = buf[j];
      }
      i += 64;
    }

    try {
      FileOutputStream fos = new FileOutputStream(DIR + node.childrenFilePath);
//      for (Node child : node.children) {
//        fos.write(nodeToByte(child));
//      }

//      for (int j = 0; j < node.children.size(); j++) {
//        fos.write(byteChildren[j]);
//      }

      fos.write(byteChildren);

      fos.close();
    } catch (IOException e) {
      return false;
    }

    return true;

  }

  public TreeSet<Node> readChildren(String childrenFilePath) {
    TreeSet<Node> children = new TreeSet<>(comparator);

    byte[] cb = readFile(childrenFilePath);
    ByteBuffer buf = ByteBuffer.wrap(cb);

    try {
      byte[] nb = new byte[64];
      int off = 0;
      while (true) {
        buf.get(nb, 0, 64);
        off += 64;
        buf.position(off);

        children.add(byteToNode(nb));
      }
    } catch (BufferUnderflowException e) {
      return children;
    }
  }

  private Node byteToNode(byte[] b) {
    ByteBuffer buf = ByteBuffer.wrap(b);

    int xl = buf.getInt();
    int xh = buf.getInt();
    int yl = buf.getInt();
    int yh = buf.getInt();
    int lhv = buf.getInt();
    short is = buf.getShort();
    boolean isEntry = is == 2;
    boolean isLeaf = is == 1 || isEntry;

    byte[] sb = new byte[42];
    buf.get(sb, 0, 42);
    String childrenFile = new String(sb, StandardCharsets.UTF_8);

    Node node = new Node(xl, yl, xh, yh, isLeaf);
    node.childrenFilePath = childrenFile.replace("\0", "");
    node.lhv = lhv;
    node.isEntry = isEntry;

    return node;
  }

  private byte[] nodeToByte(Node node) {
//    byte[] b = new byte[64];
    ByteBuffer buf = ByteBuffer.allocate(64);

    buf.putInt(node.xl);
    buf.putInt(node.xh);
    buf.putInt(node.yl);
    buf.putInt(node.yh);
    buf.putInt(node.lhv);
    if (node.isEntry) {
      buf.putShort((short) 2);
    } else if (node.isLeaf) {
      buf.putShort((short) 1);
    } else {
      buf.putShort((short) 0);
    }
//    buf.putShort(node.isLeaf ? (short) 1 : (short) 0);
//    System.out.println(node.childrenFilePath.getBytes().length);
//    System.out.println(buf);
    byte[] cfp = node.childrenFilePath.getBytes();
    buf.put(cfp, 0, cfp.length);

    return buf.array();
  }

  public boolean writeRoot(Node root) {
    try {
      FileOutputStream fos = new FileOutputStream(DIR + "_root_");
      fos.write(nodeToByte(root));
      fos.close();
    } catch (IOException e) {
      return false;
    }

    return true;
  }

  private String readFileAsString(String pathname) {
    return new String(readFile(pathname), StandardCharsets.UTF_8);
  }

  private byte[] readFile(String pathname) {
    try {
      return Files.readAllBytes(Paths.get(DIR + pathname));
    } catch (IOException e) {
      return new byte[0];
    }
  }
}
