package no.vestein.rtree.pre;

import no.vestein.rtree.lsm.BitMagic;

import java.util.*;


public class ZRTree {
  public final int maxEntries;
  private Node root;


  public ZRTree(final int maxEntries) {
    this.maxEntries = maxEntries;
    root = new Node(Integer.MAX_VALUE, Integer.MAX_VALUE, Integer.MIN_VALUE, Integer.MIN_VALUE);
    root.isEntry = false;
    root.childrenFilePath = Integer.toString(BitMagic.COUNTER.incrementAndGet());
  }

  public void insert(final Node entry) {

    // Find leaf
    Node leaf = chooseLeaf(entry);

    // Insert into leaf node
    List<Node> shouldBeUpdated = new ArrayList<>();

    if (leaf.children.size() < maxEntries) {
      //leaf.children.add(entry);
      leaf.addChild(entry);
    } else {
      Node extraNode = handleOverflow(leaf, entry);

      if (leaf.equals(root)) {
        root = new Node(false);
        root.setData(leaf);
        root.addChild(leaf);
        root.addChild(extraNode);
      } else {
        leaf.parent.children.add(extraNode);
      }
      shouldBeUpdated.add(extraNode);
    }

    shouldBeUpdated.add(leaf);

    if (!leaf.equals(root)) {
      adjustTree1(shouldBeUpdated, leaf.parent);
    }


  }

  private void adjustTree1(List<Node> nodeSet, Node parent) {
    boolean parentUpdated = false;
    for (Node node : nodeSet) {
      if (parent.updateDataIff(node)) {
        parentUpdated = true;
      }
    }

    if (parent.children.size() >= maxEntries) {
      Node p2 = handleOverflow(parent, null);
      List<Node> updated = new ArrayList<>();
      updated.add(p2);
      updated.add(parent);

      if (parent.equals(root)) {
        root = new Node(false);
        root.setData(parent);
        root.addChild(parent);
        root.addChild(p2);
      } else {
        parent.parent.addChild(p2);
        adjustTree1(updated, parent.parent);
      }
    } else {
      if (parentUpdated && !parent.equals(root)) {
        adjustTree1(Arrays.asList(parent), parent.parent);
      }
    }
  }

  private Node handleOverflow(Node leaf, Node entry) {
    Node extraNode = new Node(leaf.isLeaf);
    extraNode.neighbor = leaf;
    extraNode.parent = leaf.parent;

    List<Node> children = new ArrayList<>();
    children.addAll(leaf.children);
    if (entry != null) {
      children.add(entry);
    }
    children.sort(IO.comparator);

    leaf.children.clear();

    extraNode.setData(children.get(0));
    for (int i = 0; i < children.size() / 2; i++) {
      extraNode.addChild(children.get(i));
    }
    leaf.setData(children.get(children.size() / 2));
    for (int i = children.size() / 2; i < children.size(); i++) {
      leaf.addChild(children.get(i));
    }

    return extraNode;
  }

  private Node chooseLeaf(Node entry) {
    Node n = root;
    if (n.isLeaf) return n;

    Queue<Node> queue = new ArrayDeque<>();

    queue.add(n);

    while (queue.peek() != null) {
      Node next = queue.poll();

      if (next.isLeaf) {
        return next;
      }

      for (Node child : next.children) {
        if (child.lhv > entry.lhv) {
          queue.add(child);
          break;
        }
      }


      // Add last child to queue since no child had higher lhv than the new entry

      //queue.add(next.children.get(next.children.size() - 1));
      queue.add(next.children.last());
    }

    return queue.poll();
  }

  public Node getRoot() {
    return root;
  }

}
